import io
import struct
from collections import OrderedDict

CHAR = 'c'
SIGNED_CHAR = 'b'
UCHAR = 'B'
BOOL = '?'
SHORT = 'h'
USHORT = 'H'
INT = 'i'
UINT = 'I'
LONG = 'l'
ULONG = 'L'
LONGLONG = 'q'
ULONGLONG = 'Q'
FLOAT = 'f'
DOUBLE = 'd'
VOIDPTR = 'P'


class CStruct(type):
    def __prepare__(self, name):
        # Little Hack :D
        # Before python 3.6(PEP 520), namespace is initialised as an empty dict,
        # but for our purpose we must use a ordered dict because we must preserve
        # attribute definition order!
        return OrderedDict()

    def __new__(mcs, name, bases, dct):
        __structure__ = {}
        fmt = str()

        # Fields
        for key, value in dct.items():
            if key.startswith("_") \
                    or key.startswith("__") \
                    or key.endswith("_") \
                    or key.endswith("__") \
                    or not isinstance(value, str):
                continue
            __structure__[key] = value
            fmt += value

        # Special property and methods
        dct["__bytes__"] = CStruct.to_bytes
        dct["__parse_endianness"] = CStruct.__parse_endianness
        dct["__structure__"] = __structure__
        dct["__size__"] = struct.calcsize(fmt)
        dct["pack"] = CStruct.pack
        dct["set_endianness"] = CStruct.set_endianness
        dct["unpack"] = CStruct.unpack
        dct["unpack_from_io"] = CStruct.unpack_from_io

        if "__endianness__" not in dct:
            dct["__endianness__"] = '@'

        ist = type.__new__(mcs, name, bases, dct)
        ist.__endianness__ = CStruct.__parse_endianness(ist, ist.__endianness__)
        return ist

    @staticmethod
    def to_bytes(cls):
        return bytes(cls.pack())

    @staticmethod
    def pack(cls, endianness=None):
        structure = getattr(cls, "__structure__")
        package = bytearray()
        endianness = CStruct.__parse_endianness(cls, endianness)
        for key, value in structure.items():
            fmt = endianness + value
            package += struct.pack(fmt, getattr(cls, key))
        return package

    @staticmethod
    def unpack_from_io(cls, stream: io.RawIOBase, endianness=None):
        cls.unpack(stream.read(cls.__size__), endianness)

    @staticmethod
    def set_endianness(cls, endianness):
        cls.__endianness__ = CStruct.__parse_endianness(cls, endianness)

    @staticmethod
    def __parse_endianness(cls, endianness=None):
        if endianness is None:
            return cls.__endianness__
        if endianness == "big":
            endianness = '>'
        elif endianness == "little":
            endianness = '<'
        elif endianness != '>' and endianness != '<' and endianness != '!' and endianness != '@':
            raise ValueError("invalid argument, endianness must be one of these: @, little(<), big(>, !)")
        return endianness

    @staticmethod
    def unpack(cls, data, endianness=None):
        structure = getattr(cls, "__structure__")
        cursor = 0
        endianness = CStruct.__parse_endianness(cls, endianness)
        for key in structure:
            fmt = endianness + structure[key]
            try:
                size = struct.calcsize(fmt)
                extr = struct.unpack(fmt, data[cursor:cursor + size])
                setattr(cls, key, extr[0])
                cursor += size
            except struct.error as err:
                raise struct.error("field %s: %s" % (key, err))


def unpack_type(dtype, data, endianness=None):
    if endianness is not None:
        dtype = endianness + dtype
    return struct.unpack(dtype, data)[0]


def sizeof(dtype):
    return struct.calcsize(dtype)



import io


class MiningResult(object):
    def __init__(self):
        self.name = str()
        self.cave_begin = 0
        self.cave_end = 0
        self.cave_size = 0
        self.virtaddr = 0
        self.info = None

    def __str__(self):
        return "\n".join(["Section name:       {name}",
                          "Cave begin:         {cave_begin} - {cave_begin:#x}",
                          "Cave end:           {cave_end} - {cave_end:#x}",
                          "Cave size:          {cave_size} bytes",
                          "Virtaddr:           {virtaddr:#x}",
                          "info:               {info}"]).format(**self.__dict__)


def search4cave(stream: io.RawIOBase, section_name: str, section_size: int, section_info, cave_size: int, virtaddr: int,
                _bytes: bytes):
    caves = []
    byte_count = 0

    base = stream.tell()
    offset = 0

    while section_size > 0:
        rb = stream.read(1)
        section_size -= 1
        offset += 1

        if _bytes not in rb:
            if byte_count >= cave_size:
                mr = MiningResult()
                mr.name = section_name
                mr.cave_begin = (base + offset) - byte_count - 1
                mr.cave_end = (base + offset) - 1
                mr.cave_size = byte_count
                mr.virtaddr = virtaddr + offset - byte_count - 1
                mr.info = section_info
                caves.append(mr)
            byte_count = 0
            continue
        byte_count += 1

    stream.seek(base)
    return caves


def verifycave(stream: io.RawIOBase, cave_size, _byte: bytes):
    base = stream.tell()
    success = True
    while cave_size > 0:
        cave_size -= 1
        rb = stream.read(1)
        if _byte not in rb:
            success = False
            break
    stream.seek(base)
    return success



from cavefinder.support.cstruct import *

# *****************
# *  Fat Header   *
# *****************
# | Mach-O Header |
# *****************
# | Load Commands |
# *****************
# |     Data      |
# *****************

# DataType
MACHO_UINT32 = 4
MACHO_UINT64 = 8
MACHO_CPUTYPE = 4
MACHO_CPUSTYPE = 4
MACHO_VM_PROT = 4

# MAGICS
MACHO_MAGIC32 = 0xFEEDFACE
MACHO_CIGAM32 = 0xCEFAEDFE
MACHO_MAGIC64 = 0xFEEDFACF
MACHO_CIGAM64 = 0xCFFAEDFE

MACHO_NAMESIZE = 16

# *****************************************
# * HEADER                                *
# *****************************************
MH_FLAGS_NOUNDEFS = 0x1
MH_FLAGS_INCRLINK = 0x2
MH_FLAGS_DYLDLINK = 0x4
MH_FLAGS_BINDATLOAD = 0x8
MH_FLAGS_PREBOUND = 0x10
MH_FLAGS_SPLIT_SEGS = 0x20
MH_FLAGS_LAZY_INIT = 0x40
MH_FLAGS_TWOLEVEL = 0x80
MH_FLAGS_FORCE_FLAT = 0x100
MH_FLAGS_NOMULTIDEFS = 0x200
MH_FLAGS_NOFIXPREBINDING = 0x400
MH_FLAGS_PREBINDABLE = 0x800
MH_FLAGS_ALLMODSBOUND = 0x1000
MH_FLAGS_SUBSECTIONS_VIA_SYMBOLS = 0x2000
MH_FLAGS_CANONICAL = 0x4000
MH_FLAGS_WEAK_DEFINES = 0x8000
MH_FLAGS_BINDS_TO_WEAK = 0x10000
MH_FLAGS_ALLOW_STACK_EXECUTION = 0x20000
MH_FLAGS_ROOT_SAFE = 0x40000
MH_FLAGS_SETUID_SAFE = 0x80000
MH_FLAGS_NO_REEXPORTED_DYLIBS = 0x100000
MH_FLAGS_PIE = 0x200000
MH_FLAGS_DEAD_STRIPPABLE_DYLIB = 0x400000
MH_FLAGS_HAS_TLV_DESCRIPTORS = 0x800000
MH_FLAGS_NO_HEAP_EXECUTION = 0x1000000


class MachOHeader(object):
    ABI64 = 0x01000000
    magic = 0
    cputype = 0
    cpusubtype = 0
    filetype = 0
    ncmds = 0
    sizeofcmds = 0
    flags = 0
    reserved = 0

    def __str__(self):
        return '\n'.join(['Mach-O Header',
                          'Magic:                       0x%02x',
                          'CPU type:                    %s',
                          'CPU subtype:                 {cpusubtype:#x}',
                          'Filetype:                    %s',
                          'Number of commands:          {ncmds}',
                          'Size of commands:            {sizeofcmds} bytes',
                          'Flags:                       {flags:#x} %s']) \
                   .format(**self.__dict__) % (self.magic, self.cputype_str(), self.filetype_str(), self.flags_str())

    @property
    def endianness(self):
        return "little" if (self.magic == MACHO_MAGIC32 or self.magic == MACHO_MAGIC64) else "big"

    @property
    def wordsz(self):
        if self.magic == MACHO_MAGIC32 or self.magic == MACHO_CIGAM32:
            return 32
        elif self.magic == MACHO_MAGIC64 or self.magic == MACHO_CIGAM64:
            return 64

    def cputype_str(self):
        val = {1: "VAX",
               6: "MC680x0",
               7: "Intel x86",
               7 | MachOHeader.ABI64: "AMD x86_64",
               10: "MC98000",
               11: "HPPA",
               12: "ARM",
               13: "MC88000",
               14: "SPARC",
               15: "I860",
               18: "PowerPC",
               18 | MachOHeader.ABI64: "PowerPC64"}

        return "Unknown: %02x" % self.cputype if self.cputype not in val else val[self.cputype]

    def filetype_str(self):
        val = {0x01: "Object",
               0x02: "Execute",
               0x03: "FVMLIB",
               0x04: "CORE",
               0x05: "PRELOAD",
               0x06: "DYLIB",
               0x07: "DYLINKER",
               0x08: "Bundle",
               0x09: "DYLIB_STUB",
               0x0A: "DSYM",
               0x0B: "Kext_Bundle"}

        return "Unknown: %02x" % self.cputype if self.cputype not in val else val[self.cputype]

    def flags_str(self):
        retval = []
        val = {MH_FLAGS_NOUNDEFS: "NOUNDEFS",
               MH_FLAGS_INCRLINK: "INCRLINK",
               MH_FLAGS_DYLDLINK: "DYLDLINK",
               MH_FLAGS_BINDATLOAD: "BINDATLOAD",
               MH_FLAGS_PREBOUND: "PREBOUND",
               MH_FLAGS_SPLIT_SEGS: "SPLIT_SEGS",
               MH_FLAGS_LAZY_INIT: "LAZY_INIT",
               MH_FLAGS_TWOLEVEL: "TWOLEVEL",
               MH_FLAGS_FORCE_FLAT: "FORCE_FLAT",
               MH_FLAGS_NOMULTIDEFS: "NOMULTIDEFS",
               MH_FLAGS_NOFIXPREBINDING: "NOFIXPREBINDING",
               MH_FLAGS_PREBINDABLE: "PREBINDABLE",
               MH_FLAGS_ALLMODSBOUND: "ALLMODSBOUND",
               MH_FLAGS_SUBSECTIONS_VIA_SYMBOLS: "SUBSECTIONS_VIA_SYMBOLS",
               MH_FLAGS_CANONICAL: "CANONICAL",
               MH_FLAGS_WEAK_DEFINES: "WEAK_DEFINES",
               MH_FLAGS_BINDS_TO_WEAK: "BINDS_TO_WEAK",
               MH_FLAGS_ALLOW_STACK_EXECUTION: "ALLOW_STACK_EXECUTION",
               MH_FLAGS_ROOT_SAFE: "ROOT_SAFE",
               MH_FLAGS_SETUID_SAFE: "SETUID_SAFE",
               MH_FLAGS_NO_REEXPORTED_DYLIBS: "NO_REEXPORTED_DYLIBS",
               MH_FLAGS_PIE: "PIE",
               MH_FLAGS_DEAD_STRIPPABLE_DYLIB: "DEAD_STRIPPABLE_DYLIB",
               MH_FLAGS_HAS_TLV_DESCRIPTORS: "HAS_TLV_DESCRIPTORS",
               MH_FLAGS_NO_HEAP_EXECUTION: "NO_HEAP_EXECUTION"}

        for key in val:
            if self.flags & key == key:
                retval.append(val[key])
        return " | ".join(retval)


class MachOHeader32(MachOHeader, metaclass=CStruct):
    magic = UINT
    cputype = UINT
    cpusubtype = UINT
    filetype = UINT
    ncmds = UINT
    sizeofcmds = UINT
    flags = UINT

    def __init__(self, stream: io.RawIOBase, endianness):
        self.unpack_from_io(stream, endianness)


class MachOHeader64(MachOHeader, metaclass=CStruct):
    magic = UINT
    cputype = UINT
    cpusubtype = UINT
    filetype = UINT
    ncmds = UINT
    sizeofcmds = UINT
    flags = UINT
    reserved = UINT

    def __init__(self, stream: io.RawIOBase, endianness):
        self.unpack_from_io(stream, endianness)


LC_SEGMENT = 0x1  # segment of this file to be mapped
LC_SYMTAB = 0x2  # link-edit stab symbol table info
LC_SYMSEG = 0x3  # link-edit gdb symbol table info (obsolete)
LC_THREAD = 0x4  # thread
LC_UNIXTHREAD = 0x5  # unix thread (includes a stack)
LC_LOADFVMLIB = 0x6  # load a specified fixed VM shared library
LC_IDFVMLIB = 0x7  # fixed VM shared library identification
LC_IDENT = 0x8  # object identification info (obsolete)
LC_FVMFILE = 0x9  # fixed VM file inclusion (internal use)
LC_PREPAGE = 0xA  # prepage command (internal use)
LC_DYSYMTAB = 0xB  # dynamic link-edit symbol table info
LC_LOAD_DYLIB = 0xC  # load a dynamically linked shared library
LC_ID_DYLIB = 0xD  # dynamically linked shared lib ident
LC_LOAD_DYLINKER = 0xE  # load a dynamic linker
LC_ID_DYLINKER = 0xF  # dynamic linker identification
LC_PREBOUND_DYLIB = 0x10  # modules prebound for a dynamically
LC_ROUTINES = 0x11  # image routines
LC_SUB_FRAMEWORK = 0x12  # sub framework
LC_SUB_UMBRELLA = 0x13  # sub umbrella
LC_SUB_CLIENT = 0x14  # sub client
LC_SUB_LIBRARY = 0x15  # sub library
LC_TWOLEVEL_HINTS = 0x16  # two-level namespace lookup hints
LC_PREBIND_CKSUM = 0x17  # prebind checksum
LC_SEGMENT_64 = 0x19

VMPROT_READ = 0x01
VMPROT_WRITE = 0x02
VMPROT_EXECUTE = 0x04


class MachOCommand(object, metaclass=CStruct):
    cmd = UINT
    cmdsize = UINT

    def __init__(self, stream: io.RawIOBase, header: MachOHeader, command=None):
        if command is None:
            self.unpack_from_io(stream, header.endianness)
        else:
            self.cmd = command.cmd
            self.cmdsize = command.cmdsize

    def __str__(self):
        return "\n".join(["Mach-O Command",
                          "Command:         {cmd}",
                          "Command size:    {cmdsize}"]).format(**self.__dict__)


class MachOSegment(MachOCommand):
    segname = str()
    vmaddr = 0
    vmsize = 0
    fileoff = 0
    filesize = 0
    maxprot = 0
    initprot = 0
    nsects = 0
    flags = 0

    def __init__(self, stream: io.RawIOBase, header: MachOHeader, command: MachOCommand):
        super().__init__(stream, header, command)
        self.sections = []
        self.unpack_from_io(stream, header.endianness)

        if self.cmd != LC_SEGMENT and self.cmd != LC_SEGMENT_64:
            raise TypeError("it is not a MachOSegment")

        # Load sections
        for section in range(self.nsects):
            if header.wordsz == 32:
                self.sections.append(MachOSection32(stream, header))
            elif header.wordsz == 64:
                self.sections.append(MachOSection64(stream, header))

    @property
    def initprot_str(self):
        return MachOSegment.__vmprot_str(self.initprot)

    @property
    def maxprot_str(self):
        return MachOSegment.__vmprot_str(self.maxprot)

    @staticmethod
    def __vmprot_str(vprot):
        retval = []
        val = {VMPROT_READ: "READ",
               VMPROT_WRITE: "WRITE",
               VMPROT_EXECUTE: "EXECUTE"}
        for key in val:
            if vprot & key == key:
                retval.append(val[key])
        return " | ".join(retval)


class MachOSegment32(MachOSegment):
    segname = "16s"
    vmaddr = UINT
    vmsize = UINT
    fileoff = UINT
    filesize = UINT
    maxprot = UINT
    initprot = UINT
    nsects = UINT
    flags = UINT


class MachOSegment64(MachOSegment):
    segname = "16s"
    vmaddr = ULONGLONG
    vmsize = ULONGLONG
    fileoff = ULONGLONG
    filesize = ULONGLONG
    maxprot = UINT
    initprot = UINT
    nsects = UINT
    flags = UINT


class MachOSection(object, metaclass=CStruct):
    sectname = str()
    segname = str()
    addr = 0
    size = 0
    offset = 0
    align = 0
    reloff = 0
    nreloc = 0
    flags = 0
    reserved1 = 0
    reserved2 = 0
    reserved3 = 0

    def __init__(self, stream: io.RawIOBase, header: MachOHeader):
        self.unpack_from_io(stream, header.endianness)
        self.sectname = self.sectname.decode("ascii")
        self.segname = self.segname.decode("ascii")


class MachOSection32(MachOSection):
    sectname = "16s"
    segname = "16s"
    addr = UINT
    size = UINT
    offset = UINT
    align = UINT
    reloff = UINT
    nreloc = UINT
    flags = UINT
    reserved1 = UINT
    reserved2 = UINT


class MachOSection64(MachOSection):
    sectname = "16s"
    segname = "16s"
    addr = ULONGLONG
    size = ULONGLONG
    offset = UINT
    align = UINT
    reloff = UINT
    nreloc = UINT
    flags = UINT
    reserved1 = UINT
    reserved2 = UINT
    reserved3 = UINT


class MachO(object):
    def __init__(self, stream: io.RawIOBase):
        self.segments = []

        # Verify MachO and select wordsz and endianness
        info = unpack_type(UINT, stream.read(sizeof(UINT)))
        stream.seek(stream.tell() - sizeof(UINT))
        if info == MACHO_MAGIC32 or info == MACHO_CIGAM32:
            self.header = MachOHeader32(stream, MachO.__parse_endianness(info))
        elif info == MACHO_MAGIC64 or info == MACHO_CIGAM64:
            self.header = MachOHeader64(stream, MachO.__parse_endianness(info))
        else:
            raise TypeError("Not a valid MachO")

        self.__load_commands(stream)

    def __str__(self):
        return str(self.header)

    def __load_commands(self, stream):
        seek = stream.tell()
        for _ in range(self.header.ncmds):
            command = MachOCommand(stream, self.header)
            if command.cmd == LC_SEGMENT or command.cmd == LC_SEGMENT_64:
                if self.header.wordsz == 32:
                    segment = MachOSegment32(stream, self.header, command)
                else:
                    segment = MachOSegment64(stream, self.header, command)
                self.segments.append(segment)
            seek += command.cmdsize
            stream.seek(seek)

    @staticmethod
    def __parse_endianness(value):
        return "little" if (value == MACHO_MAGIC32 or value == MACHO_MAGIC64) else "big"

    @staticmethod
    def verify(file: io.RawIOBase):
        s_pos = file.tell()
        magic = unpack_type(UINT, file.read(sizeof(UINT)))
        file.seek(s_pos)
        return magic == MACHO_MAGIC32 or magic == MACHO_CIGAM32 or magic == MACHO_MAGIC64 or magic == MACHO_CIGAM64



import argparse
import os
import sys

from cavefinder.finder import search4cave, verifycave
from cavefinder.support.elf import *
from cavefinder.support.macho import *
from cavefinder.support.mspe import *

__version__ = "1.0.0"

WELCOME = """
                      /=============\\
                     /      | |      \\  
   ______                   | |     ______ _             __           
  / ____/____ _ _   __ ___  | |    / ____/(_)____   ____/ /___   _____
 / /    / __ `/| | / // _ \ | |   / /_   / // __ \ / __  // _ \ / ___/
/ /___ / /_/ / | |/ //  __/ | |  / __/  / // / / // /_/ //  __// /    
\____/ \__,_/  |___/ \___/  |_| /_/    /_//_/ /_/ \__,_/ \___//_/ v:%s"""


def main():
    parser = argparse.ArgumentParser(description="Dig in a binary to find all code caves")
    parser.add_argument("--size", help="minimum size of a code cave, default: 100", type=int, default=100)
    parser.add_argument("--byte", help="byte to search, default: 0x00", type=str, default="0x00")
    parser.add_argument("--payload", help="file with payload to be injected", type=str, metavar="<file_name>")
    parser.add_argument("--addr", help="address where to inject the payload", type=str, metavar="<address>", default=0)
    parser.add_argument("binary", help="executable file")
    args = parser.parse_args()

    print(WELCOME % __version__, end='\n\n')
    print("[*] Loading binary '%s'..." % args.binary, end="\n\n")

    stream = open_file(args.binary, args.payload is not None)
    btype = load_binary(stream)

    if btype is None:
        print("Unsupported binary type")
        exit(-1)

    print(btype, end="\n\n")

    if args.payload is None:
        caves = search_by_type(stream, args.size, bytes([int(args.byte.encode("ascii"), 16)]), btype)
        print("[!] Caves found: %d" % len(caves), end="\n\n")
        for cave in caves:
            print(cave, end="\n\n")
        print("[*] Mining finished")
    else:
        addr_base = int(args.addr.encode("ascii"), 16)
        payload = open_file(args.payload)
        paysize = os.path.getsize(args.payload)
        print("[*] Injecting %s (size %d bytes) into %s..." % (args.payload, paysize, args.binary))

        stream.seek(addr_base)
        if not verifycave(stream, paysize, bytes([int(args.byte.encode("ascii"), 16)])):
            print("Payload is too big for %s@0x%02x, aborted!" % (args.binary, addr_base), file=sys.stderr)
            exit(-1)

        stream.write(payload.read())
        payload.close()
        print("[*] Finished")

    stream.close()


def open_file(file: str, writable=False) -> io.RawIOBase:
    stream: io.RawIOBase = None
    try:
        stream = open(file, "rb" if not writable else "r+b")
    except FileNotFoundError as err:
        print(err, file=sys.stderr)
        exit(-1)
    return stream


def load_binary(stream):
    if Elf.verify(stream):
        return Elf(stream)
    elif MachO.verify(stream):
        return MachO(stream)
    elif Pe.verify(stream):
        return Pe(stream)
    return None


def search_by_type(stream, cave_size, _bytes, btype):
    caves = []
    if isinstance(btype, Elf):
        elf: Elf = btype
        for section in elf.sections:
            stream.seek(section.sh_offset)
            info = "Type: %s, Flags: %s" % (section.type_str(), section.flags_str())
            caves += search4cave(stream, elf.get_section_name(section), section.sh_size, info, cave_size,
                                 section.sh_addr, _bytes)
    elif isinstance(btype, MachO):
        macho: MachO = btype
        for segment in macho.segments:
            for section in segment.sections:
                stream.seek(section.offset)
                info = "%s [%s]" % (segment.initprot_str, segment.maxprot_str)
                caves += search4cave(stream, "%s.%s" % (section.segname, section.sectname), section.size, info,
                                     cave_size, section.addr, _bytes)
    elif isinstance(btype, Pe):
        pe: Pe = btype
        image_base = pe.pe_header.optional_header.image_base
        for section in pe.sections:
            stream.seek(section.ptr_rawdata)
            caves += search4cave(stream, section.name, section.size_rawdata, None, cave_size,
                                 image_base + section.virtual_addr, _bytes)

    return caves


if __name__ == "__main__":
    main()



from cavefinder.support.cstruct import *

# ---------------------
# | Dos Header        |
# ---------------------
# | Pe Signature      |
# ---------------------
# | COFF Header       |
# *********************
# | Optional Header   |
# *********************
# | Section Table     |
# ---------------------
# | Mappable sections |
# ---------------------

# 
MZ_MAGIC = 0x4D5A
MZ_CIGAM = 0x5A4D

PE_MAGIC = 0x50450000
PE_CIGAM = 0x00004550


class DosHeader(object, metaclass=CStruct):
    e_magic = USHORT  # 00: MZ Header signature
    e_cblp = USHORT  # 02: Bytes on last page of file
    e_cp = USHORT  # 04: Pages in file
    e_crlc = USHORT  # 06: Relocations
    e_cparhdr = USHORT  # 08: Size of header in paragraphs
    e_minalloc = USHORT  # 0a: Minimum extra paragraphs needed
    e_maxalloc = USHORT  # 0c: Maximum extra paragraphs needed
    e_ss = USHORT  # 0e: Initial (relative) SS value
    e_sp = USHORT  # 10: Initial SP value
    e_csum = USHORT  # 12: Checksum
    e_ip = USHORT  # 14: Initial IP value
    e_cs = USHORT  # 16: Initial (relative) CS value
    e_lfarlc = USHORT  # 18: File address of relocation table
    e_ovno = USHORT  # 1a: Overlay number
    e_res = "8s"  # 1c: Reserved words
    e_oemid = USHORT  # 24: OEM identifier (for e_oeminfo)
    e_oeminfo = USHORT  # 26: OEM information=0 e_oemid specific
    e_res2 = "20s"  # 28: Reserved words
    e_lfanew = UINT  # 3c: Offset to extended header

    def __init__(self, stream: io.RawIOBase):
        self.e_magic = unpack_type(USHORT, stream.read(2))
        stream.seek(stream.tell() - sizeof(USHORT))

        if self.e_magic == MZ_CIGAM or MZ_MAGIC:
            self.unpack_from_io(stream)
        else:
            raise TypeError("Not a valid PE (Invalid DosHeader)")

    def __str__(self):
        return '\n'.join(["Dos Header:",
                          "Magic:                       {e_magic:#x}",
                          "Byte on last page:           {e_cblp:#x}",
                          "Page in file:                {e_cp:#x}",
                          "Relocations:                 {e_crlc:#x}",
                          "Size of header:              {e_cparhdr:#x}",
                          "Min alloc:                   {e_minalloc:#x}",
                          "Max alloc:                   {e_maxalloc:#x}",
                          "ss:                          {e_ss:#x}",
                          "sp:                          {e_sp:#x}",
                          "Checksum:                    {e_csum:#x}",
                          "ip:                          {e_ip:#x}",
                          "cs:                          {e_cs:#x}",
                          "File address reloc table:    {e_lfarlc:#x}",
                          "Overlay number:              {e_ovno:#x}",
                          "OEM id:                      {e_oemid:#x}",
                          "OEM info:                    {e_oeminfo:#x}",
                          "Offset to extended header:   {e_lfanew:#x}"]).format(**self.__dict__)

    @property
    def endianness(self):
        return 'big' if self.e_magic == MZ_MAGIC else 'little'


COFF_MACHINE_UNKNOWN = 0  # unknown
COFF_MACHINE_I386 = 0x014c  # Intel 386.
COFF_MACHINE_R3000 = 0x0162  # MIPS little-endian, 0x160 big-endian
COFF_MACHINE_R4000 = 0x0166  # MIPS little-endian
COFF_MACHINE_R10000 = 0x0168  # MIPS little-endian
COFF_MACHINE_WCEMIPSV2 = 0x0169  # MIPS little-endian WCE v2
COFF_MACHINE_ALPHA = 0x0184  # Alpha_AXP
COFF_MACHINE_POWERPC = 0x01F0  # IBM PowerPC Little-Endian
COFF_MACHINE_SH3 = 0x01a2  # SH3 little-endian
COFF_MACHINE_SH3E = 0x01a4  # SH3E little-endian
COFF_MACHINE_SH4 = 0x01a6  # SH4 little-endian
COFF_MACHINE_ARM = 0x01c0  # ARM Little-Endian
COFF_MACHINE_THUMB = 0x01c2
COFF_MACHINE_AMD64 = 0x8664
COFF_MACHINE_IA64 = 0x0200  # Intel 64
COFF_MACHINE_MIPS16 = 0x0266  # MIPS
COFF_MACHINE_MIPSFPU = 0x0366  # MIPS
COFF_MACHINE_MIPSFPU16 = 0x0466  # MIPS
COFF_MACHINE_ALPHA64 = 0x0284  # ALPHA64


class COFFHeader(object, metaclass=CStruct):
    machine = USHORT
    nsections = USHORT
    timestamp = UINT
    ptr_to_symtable = UINT
    nsym = UINT
    size_opheader = USHORT
    characteristics = USHORT

    def __init__(self, stream: io.RawIOBase, endianness):
        self.unpack_from_io(stream, endianness)

    def __str__(self):
        return "\n".join(["Machine:                 {machine:#x} (%s)",
                          "Number of sections:      {nsections}",
                          "Timestamp:               {timestamp}",
                          "Ptr to symbol table:     {ptr_to_symtable:#x}",
                          "Number of symbol:        {nsym}",
                          "Size of optional header: {size_opheader}",
                          "Characteristics:         {characteristics:#x}"]).format(**self.__dict__) % self.machine_str

    @property
    def machine_str(self):
        val = {
            COFF_MACHINE_UNKNOWN: "unknown",
            COFF_MACHINE_I386: "Intel 386",
            COFF_MACHINE_R3000: "MIPS little-endian",
            COFF_MACHINE_R4000: "MIPS little-endian",
            COFF_MACHINE_R10000: "MIPS little-endian",
            COFF_MACHINE_WCEMIPSV2: "MIPS little-endian WCE v2",
            COFF_MACHINE_ALPHA: "Alpha_AXP",
            COFF_MACHINE_POWERPC: "PowerPC little-endian",
            COFF_MACHINE_SH3: "SH3 little-endian",
            COFF_MACHINE_SH3E: "SH3E little-endian",
            COFF_MACHINE_SH4: "SH4 little-endian",
            COFF_MACHINE_ARM: "ARM little-endian",
            COFF_MACHINE_THUMB: "THUMB",
            COFF_MACHINE_AMD64: "AMD 64",
            COFF_MACHINE_IA64: "Intel 64",
            COFF_MACHINE_MIPS16: "MIPS",
            COFF_MACHINE_MIPSFPU: "MIPS",
            COFF_MACHINE_MIPSFPU16: "MIPS",
            COFF_MACHINE_ALPHA64: "ALPHA64"
        }

        return "Unknown: %02x" % self.machine if self.machine not in val else val[self.machine]


IMAGE_NT_OPTIONAL_HDR32_MAGIC = 0x10b
IMAGE_NT_OPTIONAL_HDR64_MAGIC = 0x20b
IMAGE_ROM_OPTIONAL_HDR_MAGIC = 0x107

IMAGE_SUBSYSTEM_UNKNOWN = 0x0
IMAGE_SUBSYSTEM_NATIVE = 0x1
IMAGE_SUBSYSTEM_WINDOWS_GUI = 0x2
IMAGE_SUBSYSTEM_WINDOWS_CUI = 0x3
IMAGE_SUBSYSTEM_OS2_CUI = 0x5
IMAGE_SUBSYSTEM_POSIX_CUI = 0x7
IMAGE_SUBSYSTEM_NATIVE_WINDOWS = 0x8
IMAGE_SUBSYSTEM_WINDOWS_CE_GUI = 0x9


class OptionalHeader(object, metaclass=CStruct):
    magic = USHORT
    major_linker_version = UCHAR
    minor_linker_version = UCHAR
    size_of_code = UINT
    size_initialized_data = UINT
    size_uninitialized_data = UINT
    address_entry_point = UINT
    base_code = UINT
    image_base = ULONGLONG
    section_alignment = UINT
    file_alignment = UINT
    major_osversion = USHORT
    minor_osversion = USHORT
    major_image_version = USHORT
    minor_image_version = USHORT
    major_subsystem_version = USHORT
    minor_subsystem_version = USHORT
    win32_version = UINT
    size_image = UINT
    size_headers = UINT
    checksum = UINT
    subsystem = USHORT
    dll_characteristics = USHORT
    size_stack_reserve = ULONGLONG
    size_stack_commit = ULONGLONG
    size_heap_reserve = ULONGLONG
    size_heap_commit = ULONGLONG
    loader_flags = UINT
    number_rva_and_sizes = UINT

    def __init__(self, stream: io.RawIOBase, endianness):
        self.unpack_from_io(stream, endianness)

    def subsytem_str(self):
        val = {
            IMAGE_SUBSYSTEM_UNKNOWN: "Unknown",
            IMAGE_SUBSYSTEM_NATIVE: "Native",
            IMAGE_SUBSYSTEM_WINDOWS_GUI: "Windows GUI",
            IMAGE_SUBSYSTEM_WINDOWS_CUI: "Console",
            IMAGE_SUBSYSTEM_OS2_CUI: "OS/2 Console",
            IMAGE_SUBSYSTEM_POSIX_CUI: "Posix Console",
            IMAGE_SUBSYSTEM_NATIVE_WINDOWS: "diver 9x native",
            IMAGE_SUBSYSTEM_WINDOWS_CE_GUI: "Windows CE",
        }
        return "Unknown: %02x" % self.subsystem if self.subsystem not in val else val[self.subsystem]

    @property
    def wordsz(self):
        if self.magic == IMAGE_NT_OPTIONAL_HDR32_MAGIC:
            return 32
        elif self.magic == IMAGE_NT_OPTIONAL_HDR64_MAGIC:
            return 64
        return 0


class PeSectionHeader(object, metaclass=CStruct):
    name = "8s"
    physaddr_or_virtsize = UINT
    virtual_addr = UINT
    size_rawdata = UINT
    ptr_rawdata = UINT
    ptr_relocations = UINT
    ptr_linenumbers = UINT
    nrelocations = USHORT
    nlinenumbers = USHORT
    characteristics = UINT

    def __init__(self, stream: io.RawIOBase, endianness):
        self.unpack_from_io(stream, endianness)

    def __str__(self):
        return "\n".join(["Section Header",
                          "Name:                    {name}",
                          "Virtual address:         {virtual_addr:#x}",
                          "Size of raw data:        {size_rawdata:#x}",
                          "Pointer to raw data:     {ptr_rawdata:#x}",
                          "Pointer to relocations:  {ptr_relocations:#x}",
                          "Pointer to line numbers: {ptr_linenumbers:#x}",
                          "Number of relocations:   {nrelocations:#x}",
                          "Number of line numbers:  {nlinenumbers:#x}",
                          "Characteristics:         {characteristics:#x}"]).format(**self.__dict__)


class PEHeader(object):
    def __init__(self, stream: io.RawIOBase, offset=0):
        if offset > 0:
            stream.seek(offset)

        self.signature = unpack_type(UINT, stream.read(4))
        if self.signature != PE_MAGIC and self.signature != PE_CIGAM:
            raise TypeError("Not a valid PE (Invalid NT signature)")

        self.file_header = COFFHeader(stream, self.endianness)

        # Parse optional header
        self.optional_header = None
        jmp_op = stream.tell()
        if self.file_header.size_opheader > 0:
            self.optional_header = OptionalHeader(stream, self.endianness)
        stream.seek(jmp_op + self.file_header.size_opheader)

    def __str__(self):
        return str(self.file_header)

    @property
    def endianness(self):
        return "big" if self.signature == PE_MAGIC else "little"


class Pe(object):
    def __init__(self, stream: io.RawIOBase):
        self.dos_header = DosHeader(stream)
        self.pe_header = PEHeader(stream, self.dos_header.e_lfanew)

        # Parse sections
        self.sections = []
        for _ in range(self.pe_header.file_header.nsections):
            self.sections.append(PeSectionHeader(stream, self.pe_header.endianness))

    def __str__(self):
        return "\n".join(["Pe Header",
                          "Magic:                   0x%02x",
                          "%s"]) % (self.pe_header.signature, str(self.pe_header))

    @staticmethod
    def verify(file: io.RawIOBase):
        s_pos = file.tell()
        magic = unpack_type(USHORT, file.read(2))  # MS_WORD
        file.seek(s_pos)
        return magic == MZ_MAGIC or magic == MZ_CIGAM



from setuptools import setup, find_packages

setup(
    name="cavefinder",
    version="1.0.0",
    description="Codecave miner",
    long_description=open("README.md", "r").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/jacopodl/cavefinder",
    author="Jacopo De Luca",
    author_email="jacopo.delu@gmail.com",
    license="GNU General Public License v3",
    keywords=["codecave", "cave", "code", "elf", "mach-o", "portable-executable", "injection", "code-injection",
              "hacking-tool"],
    packages=find_packages(),
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Programming Language :: Python :: 3"
    ], entry_points={
        'console_scripts': ['cavefinder=cavefinder.main:main']
    })



from cavefinder.support.cstruct import *

# MAGIC
ELF_MAGIC = bytes([0x7f, 0x45, 0x4c, 0x46])

# *****************************************
# * HEADER                                *
# *****************************************
EI_NIDENT = 16
EI_MAG0 = 0x00
EI_MAG1 = 0x01
EI_MAG2 = 0x02
EI_MAG3 = 0x03
EI_CLASS = 0x04
# *****************
EI_CLASSNONE = 0x00
EI_CLASS32 = 0x01
EI_CLASS64 = 0x02
# *****************
EI_DATA = 0x05
# *****************
EI_DATANONE = 0x00
EI_DATA2LSB = 0x01
EI_DATA2MSB = 0x02
# *****************
EI_VERSION = 0x06
EI_OSABI = 0x07
EI_ABIVERSION = 0x08
EI_PAD = 0x09

# *****************************************
# * SHDR                                  *
# *****************************************
SHT_TYPE_NULL = 0x00
SHT_TYPE_PROGBITS = 0x01
SHT_TYPE_SYMTAB = 0x02
SHT_TYPE_STRTAB = 0x03
SHT_TYPE_RELA = 0x04
SHT_TYPE_HASH = 0x05
SHT_TYPE_DYNAMIC = 0x06
SHT_TYPE_NOTE = 0x07
SHT_TYPE_NOBITS = 0x08
SHT_TYPE_REL = 0x09
SHT_TYPE_SHLIB = 0x0A
SHT_TYPE_DYNSYM = 0x0B
SHT_TYPE_LOPROC = 0x70000000
SHT_TYPE_HIPROC = 0x7FFFFFFF
SHT_TYPE_LOUSER = 0x80000000
SHT_TYPE_HIUSER = 0xFFFFFFFF
# Flags
SHT_FLAGS_WRITE = 0x01
SHT_FLAGS_ALLOC = 0x02
SHT_FLAGS_EXECINSTR = 0x04
SHT_FLAGS_MASKPROC = 0xF0000000


class ElfHeader(object):
    e_ident = []
    e_type = 0
    e_machine = 0
    e_version = 0
    e_entry = 0
    e_phoff = 0
    e_shoff = 0
    e_flags = 0
    e_ehsize = 0
    e_phentsize = 0
    e_phnum = 0
    e_shentsize = 0
    e_shnum = 0
    e_shstrnd = 0

    def __str__(self):
        return '\n'.join(['ELF Header',
                          'Magic:                       %s',
                          'ABI:                         %s',
                          'Type:                        {e_type:#} (%s)',
                          'Machine:                     {e_machine:#} (%s)',
                          'Endianness and word size:    %s endian - %d bit',
                          'Version:                     {e_version}',
                          'Entry point:                 {e_entry:#x}',
                          'Program header offset:       {e_phoff} - {e_phoff:#x} (bytes in file)',
                          'Section header offset:       {e_shoff} - {e_shoff:#x} (bytes in file)',
                          'Flags:                       {e_flags:#}',
                          'Size of this header:         {e_ehsize}',
                          'Size of program headers:     {e_phentsize}',
                          'Number of program headers:   {e_phnum}',
                          'Size of section headers:     {e_shentsize}',
                          'Number of section headers:   {e_shnum}',
                          'String table index:          {e_shstrnd:#}']) \
                   .format(**self.__dict__) % (" ".join([("%02x" % x).upper() for x in self.e_ident]),
                                               self.abi,
                                               self.type_str(),
                                               self.em_str(),
                                               self.endianness,
                                               32 if self.e_ident[EI_CLASS] == EI_CLASS32 else 64)

    @property
    def abi(self):
        abi = self.e_ident[EI_OSABI]
        val = {0x00: "System V",
               0x01: "HP-UX operating system",
               0xFF: "Standalone (embedded) application"}
        return "Unknown: %02x" % abi if abi not in val else val[abi]

    @property
    def endianness(self):
        return "big" if self.e_ident[EI_DATA] == EI_DATA2MSB else "little"

    def type_str(self):
        val = {0x00: "No file type",
               0x01: "Relocatable file",
               0x02: "Executable file",
               0x03: "Shared object",
               0x04: "Core file",
               0xFF00: "Processor-specific",
               0xFFFF: "Processor-specific"}

        return "Unknown: %02x" % self.e_type if self.e_type not in val else val[self.e_type]

    def em_str(self):
        val = {0x00: "No machine",
               0x01: "AT&T WE 32100",
               0x02: "SPARC",
               0x03: "Intel 80386",
               0x04: "Motorola 68000",
               0x05: "Motorola 88000",
               0x07: "Intel 80860",
               0x08: "MIPS RS3000",
               0x14: "PowerPC",
               0x28: "ARM",
               0x2A: "Superh",
               0x3E: "AMD x86-64",
               0xB7: "AArch64"}

        return "Unknown: %02x" % self.e_machine if self.e_machine not in val else val[self.e_machine]


class ElfHeader32(ElfHeader, metaclass=CStruct):
    e_ident = "16s"
    e_type = USHORT
    e_machine = USHORT
    e_version = UINT
    e_entry = UINT
    e_phoff = UINT
    e_shoff = UINT
    e_flags = UINT
    e_ehsize = USHORT
    e_phentsize = USHORT
    e_phnum = USHORT
    e_shentsize = USHORT
    e_shnum = USHORT
    e_shstrnd = USHORT

    def __init__(self, stream: io.RawIOBase, endianness):
        self.unpack_from_io(stream, endianness)


class ElfHeader64(ElfHeader, metaclass=CStruct):
    e_ident = "16s"
    e_type = USHORT
    e_machine = USHORT
    e_version = UINT
    e_entry = ULONGLONG
    e_phoff = ULONGLONG
    e_shoff = ULONGLONG
    e_flags = UINT
    e_ehsize = USHORT
    e_phentsize = USHORT
    e_phnum = USHORT
    e_shentsize = USHORT
    e_shnum = USHORT
    e_shstrnd = USHORT

    def __init__(self, stream: io.RawIOBase, endianness):
        self.unpack_from_io(stream, endianness)


class ElfShdr(object):
    sh_name = 0
    sh_type = 0
    sh_flags = 0
    sh_addr = 0
    sh_offset = 0
    sh_size = 0
    sh_link = 0
    sh_info = 0
    sh_addralign = 0
    sh_entsize = 0

    def type_str(self):
        val = {SHT_TYPE_NULL: "SHT_NULL",
               SHT_TYPE_PROGBITS: "SHT_PROGBITS",
               SHT_TYPE_SYMTAB: "SHT_SYMTAB",
               SHT_TYPE_STRTAB: "SHT_STRTAB",
               SHT_TYPE_RELA: "SHT_RELA",
               SHT_TYPE_HASH: "SHT_HASH",
               SHT_TYPE_DYNAMIC: "SHT_DYNAMIC",
               SHT_TYPE_NOTE: "SHT_NOTE",
               SHT_TYPE_NOBITS: "SHT_NOBITS",
               SHT_TYPE_REL: "SHT_REL",
               SHT_TYPE_SHLIB: "SHT_SHLIB",
               SHT_TYPE_DYNSYM: "SHT_DYNSYM",
               SHT_TYPE_LOPROC: "SHT_LOPROC",
               SHT_TYPE_HIPROC: "SHT_HIPROC",
               SHT_TYPE_LOUSER: "SHT_LOUSER",
               SHT_TYPE_HIUSER: "SHT_HIUSER"}
        return "Unknown: %02x" % self.sh_type if self.sh_type not in val else val[self.sh_type]

    def flags_str(self):
        retval = []
        val = {SHT_FLAGS_WRITE: "SHF_WRITE",
               SHT_FLAGS_ALLOC: "SHF_ALLOC",
               SHT_FLAGS_EXECINSTR: "SHF_EXECINSTR",
               SHT_FLAGS_MASKPROC: "SHF_MASKPROC"}
        for key in val:
            if self.sh_flags & key == key:
                retval.append(val[key])
        return " | ".join(retval)


class ElfShdr32(ElfShdr, metaclass=CStruct):
    sh_name = UINT
    sh_type = UINT
    sh_flags = UINT
    sh_addr = UINT
    sh_offset = UINT
    sh_size = UINT
    sh_link = UINT
    sh_info = UINT
    sh_addralign = UINT
    sh_entsize = UINT

    def __init__(self, stream: io.RawIOBase, endianness):
        self.unpack_from_io(stream, endianness)


class ElfShdr64(ElfShdr, metaclass=CStruct):
    sh_name = UINT
    sh_type = UINT
    sh_flags = ULONGLONG
    sh_addr = ULONGLONG
    sh_offset = ULONGLONG
    sh_size = ULONGLONG
    sh_link = UINT
    sh_info = UINT
    sh_addralign = ULONGLONG
    sh_entsize = ULONGLONG

    def __init__(self, stream: io.RawIOBase, endianness):
        self.unpack_from_io(stream, endianness)


class Elf:
    def __init__(self, stream: io.RawIOBase):
        self.sections = []
        self.shstr = None
        # Verify ELF and select wordsz and endianness
        minfo = stream.read(EI_NIDENT)
        stream.seek(stream.tell() - EI_NIDENT)
        if ELF_MAGIC not in minfo[0:4]:
            raise TypeError("Not a valid ELF")

        if minfo[EI_CLASS] == EI_CLASS32:
            self.header = ElfHeader32(stream, Elf.__parse_endianness(minfo[EI_DATA]))
            self.__section_loader(stream, ElfShdr32)
        elif minfo[EI_CLASS] == EI_CLASS64:
            self.header = ElfHeader64(stream, Elf.__parse_endianness(minfo[EI_DATA]))
            self.__section_loader(stream, ElfShdr64)
        else:
            raise RuntimeError("Invalid ELF class")

        # Load strings table
        stream.seek(self.sections[self.header.e_shstrnd].sh_offset)
        self.shstr = stream.read(self.sections[self.header.e_shstrnd].sh_size)

    def __str__(self):
        return str(self.header)

    def __section_loader(self, stream: io.RawIOBase, shdr):
        seek = self.header.e_shoff
        for _ in range(self.header.e_shnum):
            stream.seek(seek)
            self.sections.append(shdr(stream, self.header.endianness))
            seek += self.header.e_shentsize

    def get_section_name(self, section: ElfShdr):
        buf = bytearray()
        cursor = section.sh_name
        byte = self.shstr[cursor]
        while byte != 0x00:
            buf.append(byte)
            cursor += 1
            byte = self.shstr[cursor]
        return buf.decode("ascii")

    @staticmethod
    def __parse_endianness(value):
        return "big" if value == EI_DATA2MSB else "little"

    @staticmethod
    def verify(file: io.RawIOBase):
        s_pos = file.tell()
        ident = file.read(EI_NIDENT)
        ret = ELF_MAGIC in ident[0:4]
        file.seek(s_pos)
        return ret
